"""
  monotonic
  ~~~~~~~~~

  This module provides a ``monotonic()`` function which returns the
  value (in fractional seconds) of a clock which never goes backwards.

  On Python 3.3 or newer, ``monotonic`` will be an alias of
  ``time.monotonic`` from the standard library. On older versions,
  it will fall back to an equivalent implementation:

  +-------------+--------------------+
  | Linux, BSD  | clock_gettime(3)   |
  +-------------+--------------------+
  | Windows     | GetTickCount64     |
  +-------------+--------------------+
  | OS X        | mach_absolute_time |
  +-------------+--------------------+

  If no suitable implementation exists for the current platform,
  attempting to import this module (or to import from it) will
  cause a RuntimeError exception to be raised.


  Copyright 2014 Ori Livneh <ori@wikimedia.org>

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

"""
import ctypes
import ctypes.util
import os
import platform
import re
import sys
import time

__all__ = ('monotonic',)


def get_os_release():
    """Get the leading numeric component of the OS release."""
    return re.match('[\d.]+', platform.release()).group(0)


def compare_versions(v1, v2):
    """Compare two version strings.
    Returns True if the former is (strictly) greater than the later,
            False otherwise.
    """

    def normalize(v):
        return tuple(map(int, re.sub(r'(\.0+)*$', '', v).split('.')))

    return normalize(v1) > normalize(v2)


try:
    monotonic = time.monotonic
except AttributeError:
    try:
        if sys.platform == 'darwin':  # OS X, iOS
            # See Technical Q&A QA1398 of the Mac Developer Library:
            #  <https://developer.apple.com/library/mac/qa/qa1398/>
            libc = ctypes.CDLL('libc.dylib', use_errno=True)


            class mach_timebase_info_data_t(ctypes.Structure):
                """System timebase info. Defined in <mach/mach_time.h>."""
                _fields_ = (('numer', ctypes.c_uint32),
                            ('denom', ctypes.c_uint32))


            mach_absolute_time = libc.mach_absolute_time
            mach_absolute_time.restype = ctypes.c_uint64

            timebase = mach_timebase_info_data_t()
            libc.mach_timebase_info(ctypes.byref(timebase))
            ticks_per_second = timebase.numer / timebase.denom * 1.0e9


            def monotonic():
                """Monotonic clock, cannot go backward."""
                return mach_absolute_time() / ticks_per_second

        elif sys.platform.startswith('win32'):
            # Windows Vista / Windows Server 2008 or newer.
            GetTickCount64 = ctypes.windll.kernel32.GetTickCount64
            GetTickCount64.restype = ctypes.c_ulonglong


            def monotonic():
                """Monotonic clock, cannot go backward."""
                return GetTickCount64() / 1000.0

        else:
            try:
                clock_gettime = ctypes.CDLL(ctypes.util.find_library('c'),
                                            use_errno=True).clock_gettime
            except AttributeError:
                clock_gettime = ctypes.CDLL(ctypes.util.find_library('rt'),
                                            use_errno=True).clock_gettime


            class timespec(ctypes.Structure):
                """Time specification, as described in clock_gettime(3)."""
                _fields_ = (('tv_sec', ctypes.c_long),
                            ('tv_nsec', ctypes.c_long))


            ts = timespec()

            if sys.platform.startswith('linux'):
                if compare_versions(get_os_release(), '2.6.28') > 0:
                    CLOCK_MONOTONIC = 4  # CLOCK_MONOTONIC_RAW
                else:
                    CLOCK_MONOTONIC = 1
            elif sys.platform.startswith('freebsd'):
                CLOCK_MONOTONIC = 4
            elif 'bsd' in sys.platform:
                CLOCK_MONOTONIC = 3


            def monotonic():
                """Monotonic clock, cannot go backward."""
                if clock_gettime(CLOCK_MONOTONIC, ctypes.pointer(ts)):
                    errno = ctypes.get_errno()
                    raise OSError(errno, os.strerror(errno))
                return ts.tv_sec + ts.tv_nsec / 1.0e9

        # Perform a sanity-check.
        if monotonic() - monotonic() > 0:
            raise ValueError('monotonic() is not monotonic!')

    except Exception as err:
        raise RuntimeError('no suitable implementation for this system: %s'
                           % err)



"""
AMQP uses exceptions to handle errors:

* Any operational error (message queue not found, insufficient access rights, etc.) results in a
  channel exception.
* Any structural error (invalid argument, bad sequence of methods, etc.) results in a connection
  exception.

According to the AMQP specification, an exception closes the associated channel or connection,  and
returns a reply code and reply text to the client. However, amqpy will automatically re-open the
channel after a channel error.
"""
import struct
from collections import namedtuple

from . import compat

compat.patch()  # monkey-patch builtins.TimeoutError

method_t = namedtuple('method_t', ('class_id', 'method_id'))

__all__ = [
    'Timeout',
    'AMQPError',
    'AMQPConnectionError', 'ChannelError',
    'RecoverableConnectionError', 'IrrecoverableConnectionError',
    'RecoverableChannelError', 'IrrecoverableChannelError',
    'ConsumerCancelled', 'ContentTooLarge', 'NoConsumers',
    'ConnectionForced', 'InvalidPath', 'AccessRefused', 'NotFound',
    'ResourceLocked', 'PreconditionFailed', 'FrameError', 'FrameSyntaxError',
    'InvalidCommand', 'ChannelNotOpen', 'UnexpectedFrame', 'ResourceError',
    'NotAllowed', 'AMQPNotImplementedError', 'InternalError',
]


class Timeout(TimeoutError):
    """General AMQP operation timeout
    """
    pass


class AMQPError(Exception):
    code = 0

    def __init__(self, reply_text=None, method_type=None, method_name=None, reply_code=None,
                 channel_id=None):
        """
        :param reply_text: localized reply text
        :param method_type: method type
        :param method_name: method name
        :param reply_code: AMQP reply (exception) code
        :param channel_id: associated channel ID, if any
        :type reply_text: str or None
        :type method_type: amqpy.spec.method_t or None
        :type method_name: str or None
        :type reply_code: int or None
        :type channel_id: int or None
        """
        self.message = reply_text
        self.reply_code = reply_code or self.code
        self.reply_text = reply_text
        self.method_type = method_type
        self.method_name = method_name or ''
        if method_type and not self.method_name:
            self.method_name = METHOD_NAME_MAP.get(method_type, '')
        self.channel_id = channel_id
        super().__init__(self, reply_code, reply_text, method_type, self.method_name, channel_id)

    def __str__(self):
        if self.method:
            return '{0.method} [ch: {0.channel_id}]: ({0.reply_code}) {0.reply_text}'.format(self)
        return self.reply_text or '<AMQPError: unknown error>'

    @property
    def method(self):
        return self.method_name or self.method_type


class AMQPConnectionError(AMQPError):
    pass


class ChannelError(AMQPError):
    pass


class RecoverableChannelError(ChannelError):
    pass


class IrrecoverableChannelError(ChannelError):
    pass


class RecoverableConnectionError(AMQPConnectionError):
    pass


class IrrecoverableConnectionError(AMQPConnectionError):
    pass


class Blocked(RecoverableConnectionError):
    pass


class ConsumerCancelled(RecoverableConnectionError):
    pass


class ContentTooLarge(RecoverableChannelError):
    """The client attempted to transfer content larger than the server could accept at the present
    time. The client may retry at a later time.
    """
    code = 311


class NoConsumers(RecoverableChannelError):
    """When the exchange cannot deliver to a consumer when the immediate flag is set. As a result
    of pending data on the queue or the absence of any consumers of the queue.
    """
    code = 313


class ConnectionForced(RecoverableConnectionError):
    """An operator intervened to close the connection for some reason. The client may retry at some
    later date.
    """
    code = 320


class InvalidPath(IrrecoverableConnectionError):
    """The client tried to work with an unknown virtual host.
    """
    code = 402


class AccessRefused(IrrecoverableChannelError):
    """The client attempted to work with a server entity to which it has no access due to
    security settings.
    """
    code = 403


class NotFound(IrrecoverableChannelError):
    """The client attempted to work with a server entity that does not exist.
    """
    code = 404


class ResourceLocked(RecoverableChannelError):
    """The client attempted to work with a server entity to which it has no access because
    another client is working
    with it.
    """
    code = 405


class PreconditionFailed(IrrecoverableChannelError):
    """The client requested a method that was not allowed because some precondition failed.
    """
    code = 406


class FrameError(IrrecoverableConnectionError):
    """The sender sent a malformed frame that the recipient could not decode. This strongly implies
    a programming error in the sending peer.
    """
    code = 501


class FrameSyntaxError(IrrecoverableConnectionError):
    """The sender sent a frame that contained illegal values for one or more fields. This strongly
    implies a programming error in the sending peer.
    """
    code = 502


class InvalidCommand(IrrecoverableConnectionError):
    """The client sent an invalid sequence of frames, attempting to perform an operation that was
    considered invalid by the server. This usually implies a programming error in the client.
    """
    code = 503


class ChannelNotOpen(IrrecoverableConnectionError):
    """The client attempted to work with a channel that had not been correctly opened. This most
    likely indicates a fault in the client layer.
    """
    code = 504


class UnexpectedFrame(IrrecoverableConnectionError):
    """The peer sent a frame that was not expected, usually in the context of a content header and
    body. This strongly indicates a fault in the peer's content processing.
    """
    code = 505


class ResourceError(RecoverableConnectionError):
    """The server could not complete the method because it lacked sufficient resources. This may be
    due to the client creating too many of some type of entity.
    """
    code = 506


class NotAllowed(IrrecoverableConnectionError):
    """The client tried to work with some entity in a manner that is prohibited by the server,  due
    to security settings or by some other criteria.
    """
    code = 530


class AMQPNotImplementedError(IrrecoverableConnectionError):
    """The client tried to use functionality that is not implemented in the server.
    """
    code = 540


class InternalError(IrrecoverableConnectionError):
    """The server could not complete the method because of an internal error. The server may
    require intervention by an operator in order to resume normal operations.
    """
    code = 541


ERROR_MAP = {
    311: ContentTooLarge,
    313: NoConsumers,
    320: ConnectionForced,
    402: InvalidPath,
    403: AccessRefused,
    404: NotFound,
    405: ResourceLocked,
    406: PreconditionFailed,
    501: FrameError,
    502: FrameSyntaxError,
    503: InvalidCommand,
    504: ChannelNotOpen,
    505: UnexpectedFrame,
    506: ResourceError,
    530: NotAllowed,
    540: AMQPNotImplementedError,
    541: InternalError,
}


def error_for_code(code, text, meth_type, default, channel_id=None):
    """Get exception class associated with specified error code

    :param int code: AMQP reply code
    :param str text: localized reply text
    :param meth_type: method type
    :param default: default exception class if error code cannot be matched with an exception class
    :param channel_id: optional associated channel ID
    :type meth_type: amqpy.spec.method_t
    :type default: Callable
    :type channel_id: int or None
    :return: Exception object
    :rtype: Exception
    """
    try:
        exc = ERROR_MAP[code]
        return exc(text, meth_type, reply_code=code, channel_id=channel_id)
    except KeyError:
        return default(text, meth_type, reply_code=code, channel_id=channel_id)


METHOD_NAME_MAP = {
    method_t(10, 10): 'connection.start',
    method_t(10, 11): 'connection.start-ok',
    method_t(10, 20): 'connection.secure',
    method_t(10, 21): 'connection.secure-ok',
    method_t(10, 30): 'connection.tune',
    method_t(10, 31): 'connection.tune-ok',
    method_t(10, 40): 'connection.open',
    method_t(10, 41): 'connection.open-ok',
    method_t(10, 50): 'connection.close',
    method_t(10, 51): 'connection.close-ok',
    method_t(10, 60): 'connection.blocked',
    method_t(10, 61): 'connection.unblocked',
    method_t(20, 10): 'channel.open',
    method_t(20, 11): 'channel.open-ok',
    method_t(20, 20): 'channel.flow',
    method_t(20, 21): 'channel.flow-ok',
    method_t(20, 40): 'channel.close',
    method_t(20, 41): 'channel.close-ok',
    method_t(30, 10): 'access.request',
    method_t(30, 11): 'access.request-ok',
    method_t(40, 10): 'exchange.declare',
    method_t(40, 11): 'exchange.declare-ok',
    method_t(40, 20): 'exchange.delete',
    method_t(40, 21): 'exchange.delete-ok',
    method_t(40, 30): 'exchange.bind',
    method_t(40, 31): 'exchange.bind-ok',
    method_t(40, 40): 'exchange.unbind',
    method_t(40, 51): 'exchange.unbind-ok',
    method_t(50, 10): 'queue.declare',
    method_t(50, 11): 'queue.declare-ok',
    method_t(50, 20): 'queue.bind',
    method_t(50, 21): 'queue.bind-ok',
    method_t(50, 30): 'queue.purge',
    method_t(50, 31): 'queue.purge-ok',
    method_t(50, 40): 'queue.delete',
    method_t(50, 41): 'queue.delete-ok',
    method_t(50, 50): 'queue.unbind',
    method_t(50, 51): 'queue.unbind-ok',
    method_t(60, 10): 'basic.qos',
    method_t(60, 11): 'basic.qos-ok',
    method_t(60, 20): 'basic.consume',
    method_t(60, 21): 'basic.consume-ok',
    method_t(60, 30): 'basic.cancel',
    method_t(60, 31): 'basic.cancel-ok',
    method_t(60, 40): 'basic.publish',
    method_t(60, 50): 'basic.return',
    method_t(60, 60): 'basic.deliver',
    method_t(60, 70): 'basic.get',
    method_t(60, 71): 'basic.get-ok',
    method_t(60, 72): 'basic.get-empty',
    method_t(60, 80): 'basic.ack',
    method_t(60, 90): 'basic.reject',
    method_t(60, 100): 'basic.recover-async',
    method_t(60, 110): 'basic.recover',
    method_t(60, 111): 'basic.recover-ok',
    method_t(60, 120): 'basic.nack',
    method_t(90, 10): 'tx.select',
    method_t(90, 11): 'tx.select-ok',
    method_t(90, 20): 'tx.commit',
    method_t(90, 21): 'tx.commit-ok',
    method_t(90, 30): 'tx.rollback',
    method_t(90, 31): 'tx.rollback-ok',
    method_t(85, 10): 'confirm.select',
    method_t(85, 11): 'confirm.select-ok',
}


# insert keys which are 4-byte unsigned int representations of a method type for easy lookups
for mt, name in list(METHOD_NAME_MAP.items()):
    data = struct.pack('>HH', *mt)
    METHOD_NAME_MAP[struct.unpack('>I', data)[0]] = name



VERSION = (0, 11, 7)
__version__ = '.'.join(map(str, VERSION[0:3])) + ''.join(VERSION[3:])
__author__ = 'veegee'
__maintainer__ = 'veegee'
__contact__ = 'veegee@veegee.org'
__homepage__ = 'http://github.com/veegee/amqpy'
__docformat__ = 'restructuredtext'

from .connection import Connection
from .channel import Channel
from .message import Message
from .consumer import AbstractConsumer
from .spec import basic_return_t, queue_declare_ok_t, method_t
from .exceptions import (
    Timeout,
    AMQPError,
    AMQPConnectionError,
    RecoverableConnectionError,
    IrrecoverableConnectionError,
    ChannelError,
    RecoverableChannelError,
    IrrecoverableChannelError,
    ConsumerCancelled,
    ContentTooLarge,
    NoConsumers,
    ConnectionForced,
    InvalidPath,
    AccessRefused,
    NotFound,
    ResourceLocked,
    PreconditionFailed,
    FrameError,
    FrameSyntaxError,
    InvalidCommand,
    ChannelNotOpen,
    UnexpectedFrame,
    ResourceError,
    NotAllowed,
    AMQPNotImplementedError,
    InternalError,
    error_for_code,
    __all__ as _all_exceptions,
)

__all__ = ['Connection', 'Channel', 'Message', 'AbstractConsumer',
           'basic_return_t', 'queue_declare_ok_t', 'method_t'] + _all_exceptions



from .. import Message


class TestChannel:
    def test_with(self, conn):
        with conn as cxn:
            assert cxn.transport is not None

            with cxn.channel(1) as ch:
                assert 1 in cxn.channels

                # do something with the channel
                ch.exchange_declare('unittest.fanout', 'fanout', auto_delete=True)

                msg = Message('unittest message',
                              content_type='text/plain',
                              application_headers={'foo': 7, 'bar': 'baz'})

                ch.basic_publish(msg, 'unittest.fanout')

            # check that the channel was closed
            assert 1 not in cxn.channels
            assert ch.is_open is False

        # check that the connection was closed
        assert cxn.transport is None



from datetime import datetime
from decimal import Decimal
import pickle

from .. import Message


class TestBasicMessage:
    def check_proplist(self, msg):
        """Check roundtrip processing of a single object
        """
        raw_properties = msg.serialize_properties()

        new_msg = Message()
        new_msg.load_properties(raw_properties)
        new_msg.body = msg.body

        assert msg == new_msg

    def test_eq(self):
        msg = Message('hello', content_type='text/plain')
        assert msg

        # Make sure that something that looks vaguely like a Message doesn't raise an Attribute
        # error when compared to a Message, and instead returns False
        class FakeMsg:
            pass

        fake_msg = FakeMsg()
        fake_msg.properties = {'content_type': 'text/plain'}

        assert msg != fake_msg

    def test_pickle(self):
        msg = Message(
            'some body' * 200000,
            content_type='text/plain',
            content_encoding='utf-8',
            application_headers={'foo': 7, 'bar': 'baz', 'd2': {'foo2': 'xxx', 'foo3': -1}},
            delivery_mode=1,
            priority=7,
        )

        msg2 = pickle.loads(pickle.dumps(msg))
        assert msg == msg2

    def test_roundtrip(self):
        """Check round-trip processing of content-properties
        """
        self.check_proplist(Message())
        self.check_proplist(Message(content_type='text/plain'))
        self.check_proplist(Message(
            content_type='text/plain',
            content_encoding='utf-8',
            application_headers={'foo': 7, 'bar': 'baz', 'd2': {'foo2': 'xxx', 'foo3': -1}},
            delivery_mode=1,
            priority=7,
        ))

        self.check_proplist(Message(
            application_headers={
                'regular': datetime(2007, 11, 12, 12, 34, 56),
                'dst': datetime(2007, 7, 12, 12, 34, 56),
            },
        ))

        n = datetime.now()
        # AMQP only does timestamps to 1-second resolution
        n = n.replace(microsecond=0)
        self.check_proplist(Message(application_headers={'foo': n}))
        self.check_proplist(Message(application_headers={'foo': Decimal('10.1')}))
        self.check_proplist(Message(application_headers={'foo': Decimal('-1987654.193')}))
        self.check_proplist(Message(timestamp=datetime(1980, 1, 2, 3, 4, 6)))



from abc import ABCMeta, abstractmethod


class AbstractConsumer(metaclass=ABCMeta):
    """
    This class provides facilities to create and manage queue consumers. To
    create a consumer, subclass this class and override the :meth:`run`
    method. Then, instantiate the class with the desired parameters and call
    :meth:`declare` to declare the consumer to the server.

    Example::

        class Consumer(AbstractConsumer):
            def run(self, msg: Message):
                print('Received message: {}'.format(msg.body))
                msg.ack()

        c1 = Consumer(ch, 'test.q')
        c1.declare()

        conn.drain_events()
    """

    def __init__(self, channel, queue, consumer_tag='', no_local=False,
                 no_ack=False, exclusive=False):
        """
        :param channel: channel
        :type channel: amqpy.channel.Channel
        :param str queue: queue
        :param str consumer_tag: consumer tag, local to the connection; leave
            blank to let server auto-assign a tag
        :param bool no_local: if True: do not deliver own messages
        :param bool no_ack: server will not expect an ack for each message
        :param bool exclusive: request exclusive access
        """
        self.channel = channel
        self.queue = queue
        self.consumer_tag = consumer_tag
        self.no_local = no_local
        self.no_ack = no_ack
        self.exclusive = exclusive

        #: Number of messages consumed (incremented automatically)
        self.consume_count = 0

    def declare(self):
        """Declare the consumer

        This method calls :meth:`~amqpy.channel.Channel.basic_consume()`
        internally.

        After the queue consumer is created, :attr:`self.consumer_tag` is
        set to the server-assigned consumer tag if a tag was not specified
        initially.
        """
        self.consumer_tag = self.channel.basic_consume(
            self.queue, self.consumer_tag, self.no_local, self.no_ack, self.exclusive,
            callback=self.start, on_cancel=self.cancel_cb)

    def cancel(self):
        """Cancel the consumer
        """
        self.channel.basic_cancel(self.consumer_tag)

    @abstractmethod
    def run(self, msg):
        """Consumer callback

        This method is called when the consumer is delivered a message. This
        method **must** be overridden in the subclass.

        :param msg: received message
        :type msg: amqpy.message.Message
        """
        pass

    def cancel_cb(self, consumer_tag):
        """Consumer cancel callback

        This method is called when the consumer is cancelled. This method may
        be overridden in the subclass.

        :param str consumer_tag: consumer tag
        """
        pass

    def start(self, msg):
        self.run(msg)
        self.consume_count += 1
